% --COPYRIGHT--,BSD
% Copyright (c) 2016, Texas Instruments Incorporated
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%
% *  Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%
% *  Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in the
%    documentation and/or other materials provided with the distribution.
%
% *  Neither the name of Texas Instruments Incorporated nor the names of
%    its contributors may be used to endorse or promote products derived
%    from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
% OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
% WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
% EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
% --/COPYRIGHT--*/

%*******************************************************************************
% coefficients.m: Generate coefficients for the LEA TI design
% (TIDM-FILTERING-SIGNALPROCESSING-LEA)
%
% Brent Peterson
% Texas Instruments 2016
%*******************************************************************************

% General parameters
plotFreqz = true;
directoryName = 'coefficients';

% Create directory if it doesn't exist
mkdir(directoryName);

% Filter parameters, 20kHz sample frequency
fs = 20000;
passBand = 400;
stopBand = 400;
firOrder = 118;
iirOrder = 8;
iirStopAttn = 80;

% List of FIR filters to generate coefficients for
firFilters = [
    struct('name', 'firLowPass1kHz', 'wc', 1000/(fs/2), 'type', 'low')
    struct('name', 'firLowPass2kHz', 'wc', 2000/(fs/2), 'type', 'low')
    struct('name', 'firLowPass3kHz', 'wc', 3000/(fs/2), 'type', 'low')
    struct('name', 'firLowPass4kHz', 'wc', 4000/(fs/2), 'type', 'low')
    struct('name', 'firHighPass1kHz', 'wc', 1000/(fs/2), 'type', 'high')
    struct('name', 'firHighPass2kHz', 'wc', 2000/(fs/2), 'type', 'high')
    struct('name', 'firHighPass3kHz', 'wc', 3000/(fs/2), 'type', 'high')
    struct('name', 'firHighPass4kHz', 'wc', 4000/(fs/2), 'type', 'high')
    struct('name', 'firBandPass1kHz', 'wc', [1000-passBand 1000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'firBandPass2kHz', 'wc', [2000-passBand 2000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'firBandPass3kHz', 'wc', [3000-passBand 3000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'firBandPass4kHz', 'wc', [4000-passBand 4000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'firBandStop1kHz', 'wc', [1000-stopBand 1000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'firBandStop2kHz', 'wc', [2000-stopBand 2000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'firBandStop3kHz', 'wc', [3000-stopBand 3000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'firBandStop4kHz', 'wc', [4000-stopBand 4000+stopBand]/(fs/2), 'type', 'stop')
];

% List of IIR filters to generate coefficients for
iirFilters = [
    struct('name', 'iirLowPass1kHz', 'wc', 1000/(fs/2), 'type', 'low')
    struct('name', 'iirLowPass2kHz', 'wc', 2000/(fs/2), 'type', 'low')
    struct('name', 'iirLowPass3kHz', 'wc', 3000/(fs/2), 'type', 'low')
    struct('name', 'iirLowPass4kHz', 'wc', 4000/(fs/2), 'type', 'low')
    struct('name', 'iirHighPass1kHz', 'wc', 1000/(fs/2), 'type', 'high')
    struct('name', 'iirHighPass2kHz', 'wc', 2000/(fs/2), 'type', 'high')
    struct('name', 'iirHighPass3kHz', 'wc', 3000/(fs/2), 'type', 'high')
    struct('name', 'iirHighPass4kHz', 'wc', 4000/(fs/2), 'type', 'high')
    struct('name', 'iirBandPass1kHz', 'wc', [1000-passBand 1000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'iirBandPass2kHz', 'wc', [2000-passBand 2000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'iirBandPass3kHz', 'wc', [3000-passBand 3000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'iirBandPass4kHz', 'wc', [4000-passBand 4000+passBand]/(fs/2), 'type', 'bandpass')
    struct('name', 'iirBandStop1kHz', 'wc', [1000-stopBand 1000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'iirBandStop2kHz', 'wc', [2000-stopBand 2000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'iirBandStop3kHz', 'wc', [3000-stopBand 3000+stopBand]/(fs/2), 'type', 'stop')
    struct('name', 'iirBandStop4kHz', 'wc', [4000-stopBand 4000+stopBand]/(fs/2), 'type', 'stop')
];

% Generate coefficients from the list of FIR filters 
for i = 1:length(firFilters)
    % Generate FIR filter and pad length to even
    b = fir1(firOrder, firFilters(i).wc, firFilters(i).type);
    b = [0 b];
    
    % Plot filter frequency and phase response
    if (plotFreqz)
        figure; hold all;
        freqz(b, 1, [], fs);
        ylim([-80 20]);
    end;

    % Create and write source file
    fc = fopen(strcat(directoryName,'/',firFilters(i).name,'.c'), 'w');
    fprintf(fc, '#include "DSPLib.h"\n\n');
    fprintf(fc, ['const _q15 ',firFilters(i).name,'[',num2str(length(b)),'] = {']);
    for i = 1:length(b)
        fprintf(fc, ['\n    _Q15(',num2str(b(i),'%+1.6f'),'),']);
    end
    fprintf(fc, '\n};\n');
    fclose(fc);
end

% Generate coefficients from the list of IIR filters 
for i = 1:length(iirFilters)
    % Generate IIR filter
    [z,p,k] = cheby2(iirOrder, iirStopAttn, iirFilters(i).wc, iirFilters(i).type);
    sos = zp2sos(z, p, k);
    
    % L-inf norm scaling of SOS coefficients
    h = ones(256,1);
    g = 1.0;
    for j=1:size(sos,1)
        h = g .* h .* freqz(sos(j,1:3), sos(j,4:6), 256);
        g = 1/max(abs(h));
        sos(j,1:3) = sos(j,1:3) .* g;
    end
    
    % Plot filter frequency and phase response
    if (plotFreqz)
        figure; hold all;
        freqz(sos, [], fs);
        ylim([-80 20]);
    end;
    
    % Print IIR coefficients to C file
    fc = fopen(strcat(directoryName,'/',iirFilters(i).name,'.c'), 'w');
    fprintf(fc, '#include "DSPLib.h"\n\n');
    fprintf(fc, ['const msp_biquad_df2_q15_coeffs ',iirFilters(i).name,'[',num2str(size(sos,1)),'] = {']);
    for k = 1:size(sos,1)
        fprintf(fc, '\n    {\n');
        fprintf(fc, ['        .b0     = _Q15(',num2str(sos(k,1),'%+1.6f'),'),\n']);
        fprintf(fc, ['        .b1By2  = _Q15(',num2str(sos(k,2),'%+1.6f'),'/2),\n']);
        fprintf(fc, ['        .b2     = _Q15(',num2str(sos(k,3),'%+1.6f'),'),\n']);
        fprintf(fc, ['        .a0     = _Q15(',num2str(sos(k,4),'%+1.6f'),'),\n']);
        fprintf(fc, ['        .a1By2  = _Q15(',num2str(-sos(k,5),'%+1.6f'),'/2),\n']);
        fprintf(fc, ['        .a2     = _Q15(',num2str(-sos(k,6),'%+1.6f'),')\n']);
        fprintf(fc, '    },');
    end
    fprintf(fc, '\n};\n');
    fclose(fc);
end
