# --COPYRIGHT--,BSD
# Copyright (c) 2016, Texas Instruments Incorporated
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# *  Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# *  Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# *  Neither the name of Texas Instruments Incorporated nor the names of
#    its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# --/COPYRIGHT--*/

#!/usr/bin/env python

#*******************************************************************************
# coefficients.m: Generate coefficients for the LEA TI design
# (TIDM-FILTERING-SIGNALPROCESSING-LEA)
#
# Brent Peterson
# Texas Instruments 2016
#*******************************************************************************

import os
import numpy as np
from scipy import signal
import matplotlib.pyplot as plt

# General parameters
plotFreqz = True
directoryName = 'coefficients'

# Create directory if it doesn't exist
if not os.path.exists(directoryName):
    os.makedirs(directoryName)

# Filter parameters, 20kHz sample frequency
fs = 20000
passBand = 400
stopBand = 400
firOrder = 119
iirOrder = 8
iirStopAttn = 80

# List of FIR filters to generate coefficients for
firFilters = [
    {'name': 'firLowPass1kHz', 'order': firOrder, 'wc': 1000.0/(fs/2), 'pass_zero': True},
    {'name': 'firLowPass2kHz', 'order': firOrder, 'wc': 2000.0/(fs/2), 'pass_zero': True},
    {'name': 'firLowPass3kHz', 'order': firOrder, 'wc': 3000.0/(fs/2), 'pass_zero': True},
    {'name': 'firLowPass4kHz', 'order': firOrder, 'wc': 4000.0/(fs/2), 'pass_zero': True},
    {'name': 'firHighPass1kHz', 'order': firOrder, 'wc': 1000.0/(fs/2), 'pass_zero': False},
    {'name': 'firHighPass2kHz', 'order': firOrder, 'wc': 2000.0/(fs/2), 'pass_zero': False},
    {'name': 'firHighPass3kHz', 'order': firOrder, 'wc': 3000.0/(fs/2), 'pass_zero': False},
    {'name': 'firHighPass4kHz', 'order': firOrder, 'wc': 4000.0/(fs/2), 'pass_zero': False},
    {'name': 'firBandPass1kHz', 'order': firOrder, 'wc': np.array([1000.0-passBand, 1000.0+passBand])/(fs/2), 'pass_zero': False},
    {'name': 'firBandPass2kHz', 'order': firOrder, 'wc': np.array([2000.0-passBand, 2000.0+passBand])/(fs/2), 'pass_zero': False},
    {'name': 'firBandPass3kHz', 'order': firOrder, 'wc': np.array([3000.0-passBand, 3000.0+passBand])/(fs/2), 'pass_zero': False},
    {'name': 'firBandPass4kHz', 'order': firOrder, 'wc': np.array([4000.0-passBand, 4000.0+passBand])/(fs/2), 'pass_zero': False},
    {'name': 'firBandStop1kHz', 'order': firOrder, 'wc': np.array([1000.0-stopBand, 1000.0+stopBand])/(fs/2), 'pass_zero': True},
    {'name': 'firBandStop2kHz', 'order': firOrder, 'wc': np.array([2000.0-stopBand, 2000.0+stopBand])/(fs/2), 'pass_zero': True},
    {'name': 'firBandStop3kHz', 'order': firOrder, 'wc': np.array([3000.0-stopBand, 3000.0+stopBand])/(fs/2), 'pass_zero': True},
    {'name': 'firBandStop4kHz', 'order': firOrder, 'wc': np.array([4000.0-stopBand, 4000.0+stopBand])/(fs/2), 'pass_zero': True}
];

# List of IIR filters to generate coefficients for
iirFilters = [
    {'name': 'iirLowPass1kHz', 'order': iirOrder, 'wc': 1000.0/(fs/2), 'type': 'lowpass'},
    {'name': 'iirLowPass2kHz', 'order': iirOrder, 'wc': 2000.0/(fs/2), 'type': 'lowpass'},
    {'name': 'iirLowPass3kHz', 'order': iirOrder, 'wc': 3000.0/(fs/2), 'type': 'lowpass'},
    {'name': 'iirLowPass4kHz', 'order': iirOrder, 'wc': 4000.0/(fs/2), 'type': 'lowpass'},
    {'name': 'iirHighPass1kHz', 'order': iirOrder, 'wc': 1000.0/(fs/2), 'type': 'highpass'},
    {'name': 'iirHighPass2kHz', 'order': iirOrder, 'wc': 2000.0/(fs/2), 'type': 'highpass'},
    {'name': 'iirHighPass3kHz', 'order': iirOrder, 'wc': 3000.0/(fs/2), 'type': 'highpass'},
    {'name': 'iirHighPass4kHz', 'order': iirOrder, 'wc': 4000.0/(fs/2), 'type': 'highpass'},
    {'name': 'iirBandPass1kHz', 'order': iirOrder, 'wc': np.array([1000.0-passBand, 1000.0+passBand])/(fs/2), 'type': 'bandpass'},
    {'name': 'iirBandPass2kHz', 'order': iirOrder, 'wc': np.array([2000.0-passBand, 2000.0+passBand])/(fs/2), 'type': 'bandpass'},
    {'name': 'iirBandPass3kHz', 'order': iirOrder, 'wc': np.array([3000.0-passBand, 3000.0+passBand])/(fs/2), 'type': 'bandpass'},
    {'name': 'iirBandPass4kHz', 'order': iirOrder, 'wc': np.array([4000.0-passBand, 4000.0+passBand])/(fs/2), 'type': 'bandpass'},
    {'name': 'iirBandStop1kHz', 'order': iirOrder, 'wc': np.array([1000.0-stopBand, 1000.0+stopBand])/(fs/2), 'type': 'bandstop'},
    {'name': 'iirBandStop2kHz', 'order': iirOrder, 'wc': np.array([2000.0-stopBand, 2000.0+stopBand])/(fs/2), 'type': 'bandstop'},
    {'name': 'iirBandStop3kHz', 'order': iirOrder, 'wc': np.array([3000.0-stopBand, 3000.0+stopBand])/(fs/2), 'type': 'bandstop'},
    {'name': 'iirBandStop4kHz', 'order': iirOrder, 'wc': np.array([4000.0-stopBand, 4000.0+stopBand])/(fs/2), 'type': 'bandstop'}
];

# Generate coefficients from the list of FIR filters 
for filter in firFilters:
    # Generate FIR filter and pad length to even
    b = signal.firwin(filter['order'], filter['wc'], pass_zero=filter['pass_zero'])
    b = np.insert(b, 0, 0)
    
    # Plot filter frequency and phase response
    if (plotFreqz):
        w, h = signal.freqz(b)
        fig = plt.figure()
        plt.title('Digital filter frequency response')
        ax1 = fig.add_subplot(111)
        plt.plot(fs*w/(2*np.pi), 20 * np.log10(abs(h)), 'b')
        plt.ylabel('Amplitude [dB]', color='b')
        plt.xlabel('Frequency [Hz]')
        ax2 = ax1.twinx()
        plt.plot(fs*w/(2*np.pi), np.unwrap(np.angle(h)), 'r')
        plt.ylabel('Angle (radians)', color='r')
        plt.grid()
        plt.show()
    
    # Create and write source file
    with open('%s/%s.c' % (directoryName, filter['name']), 'w') as fc:
        fc.write('#include "DSPLib.h"\n\n')
        fc.write('const _q15 %s[%d] = {' % (filter['name'], len(b)))
        for c in b:
            fc.write('\n    _Q15(%+1.6f),' % c)
        fc.write('\n};\n')

# Generate coefficients from the list of IIR filters 
for filter in iirFilters:
    # Generate IIR filter
    sos = signal.iirfilter(filter['order'], filter['wc'], None, iirStopAttn, btype=filter['type'], analog=False, ftype='cheby2', output='sos')
    
    # L-inf norm scaling of SOS coefficients
    h = np.ones(256);
    g = 1.0;
    for i in range(len(sos)):
        h = g * h * signal.freqz(sos[i,0:3], sos[i,3:6], 256)[1];
        g = 1/np.max(np.abs(h));
        sos[i][0:3] = sos[i][0:3] * g;
    
    # Plot filter frequency and phase response
    if (plotFreqz):
        b, a = signal.sos2tf(sos)
        w, h = signal.freqz(b, a)
        fig = plt.figure()
        plt.title('Digital filter frequency response')
        ax1 = fig.add_subplot(111)
        plt.plot(fs*w/(2*np.pi), 20 * np.log10(abs(h)), 'b')
        plt.ylabel('Amplitude [dB]', color='b')
        plt.xlabel('Frequency [Hz]')
        ax2 = ax1.twinx()
        plt.plot(fs*w/(2*np.pi), np.unwrap(np.angle(h)), 'r')
        plt.ylabel('Angle (radians)', color='r')
        plt.grid()
        plt.show()
    
    # Create and write source file
    with open('%s/%s.c' % (directoryName, filter['name']), 'w') as fc:
        fc.write('#include "DSPLib.h"\n\n')
        fc.write('const msp_biquad_df2_q15_coeffs %s[%d] = {' % (filter['name'], len(sos)))
        for row in sos:
            fc.write('\n    {\n')
            fc.write('        .b0     = _Q15(%+1.6f),\n' % row[0])
            fc.write('        .b1By2  = _Q15(%+1.6f/2),\n' % row[1])
            fc.write('        .b2     = _Q15(%+1.6f),\n' % row[2])
            fc.write('        .a0     = _Q15(%+1.6f),\n' % row[3])
            fc.write('        .a1By2  = _Q15(%+1.6f/2),\n' % -row[4])
            fc.write('        .a2     = _Q15(%+1.6f)\n' % -row[5])
            fc.write('    },')
        fc.write('\n};\n')
