/* --COPYRIGHT--,BSD
 * Copyright (c) 2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
//******************************************************************************
//!  TI-Design Signal Processing with LEA on MSP430FR5994
//!
//!  William Goh, Brent Peterson, L Reynoso
//!  Texas Instruments Inc.
//!  Sept 2019
//******************************************************************************

#include "application.h"

//******************************************************************************
// Globals
//******************************************************************************
// Allocate DSPLib structures
static msp_status status;
static msp_mpy_q15_params mpyParams;
static msp_cmplx_fft_q15_params fftParams;
static msp_cmplx_fill_q15_params fillParams;
static msp_interleave_q15_params interleaveParams;

// Rectangular to clear my FFT image after every round
const Graphics_Rectangle clearSpace = { 0, 10, LCD_HORIZONTAL_MAX, LCD_VERTICAL_MAX-13 };
const Graphics_Rectangle clearSpaceTopText = { 0, 0, LCD_VERTICAL_MAX, 7 };
const Graphics_Rectangle clearSpaceCycleText = { 10, LCD_VERTICAL_MAX-11, 65, LCD_VERTICAL_MAX };


static uint8_t refreshDisp = 0;
static int8_t str[20];

extern const _q15 hammingWindow[VECTOR_SIZE];
extern void FFT_430(int16_t *input);

static void copyData(const void *src, void *dst, uint16_t length)
{
    uint16_t i;
    uint16_t *srcPtr;
    uint16_t *dstPtr;

    // Set src and dst pointers
    srcPtr = (uint16_t *)src;
    dstPtr = (uint16_t *)dst;

    for (i = 0; i < length/2; i++) {
        *dstPtr++ = *srcPtr++;
    }
}

void initAudio(void)
{
    // Initialize the microphone for recording
    gAudioConfig.audioBuffer1 = (int16_t *)&leaMemoryStartAdd.fftDataParam.samples[0];
    gAudioConfig.audioBuffer2 = (int16_t *)&leaMemoryStartAdd.fftDataParam.samples[VECTOR_SIZE];
    gAudioConfig.bufferSize = VECTOR_SIZE;
    gAudioConfig.sampleRate = FFT_SAMPLING_FREQUENCY;
    Audio_setupCollect(&gAudioConfig);

    // Start the recording by enabling the timer
    Audio_startCollect(&gAudioConfig);
}

void deinitAudio(void)
{
    // Stop the audio collection
    Audio_stopCollect(&gAudioConfig);
    Audio_shutdownCollect(&gAudioConfig);
}

void runFft(bool useLea)
{
    uint16_t i;
    int16_t *inputPtr;
    int16_t *bufferPtr;

    keepAppRunning = true;

    // Initialize the number of times to refresh the display
    refreshDisp = 2;

    // Clear the screen first
    Graphics_clearDisplay(&g_sContext);

    // Draw the basic structure
    GrStringDraw(&g_sContext, "0",
                 AUTO_STRING_LENGTH, 1, LCD_VERTICAL_MAX-10,
                 TRANSPARENT_TEXT);
    GrStringDraw(&g_sContext, "Fs 2",
                 AUTO_STRING_LENGTH, 70, LCD_VERTICAL_MAX-10,
                 TRANSPARENT_TEXT);
    GrLineDrawH(&g_sContext, 0, LCD_VERTICAL_MAX, LCD_VERTICAL_MAX-12);

    // Draw the title
    if (useLea) {
        Graphics_drawString(&g_sContext, "FFT w/ LEA",
                            AUTO_STRING_LENGTH, 0, 0,
                            GRAPHICS_OPAQUE_TEXT);
    }
    else {
        Graphics_drawString(&g_sContext, "FFT w/o LEA",
                            AUTO_STRING_LENGTH, 0, 0,
                            GRAPHICS_OPAQUE_TEXT);
    }

    // Flush Buffer to LCD
    Graphics_flushBuffer(&g_sContext);

    // Initialize audio configuration
    initAudio();

    // Initialize multiply parameters and copy Hamming window
    mpyParams.length = VECTOR_SIZE;
    copyData(hammingWindow, leaMemoryStartAdd.fftDataParam.hamming, sizeof(hammingWindow));

    // Initialize complex fill parameters
    fillParams.length = VECTOR_SIZE;
    fillParams.realValue = 0;
    fillParams.imagValue = 0;

    // Initialize interleave parameters
    interleaveParams.length = VECTOR_SIZE;
    interleaveParams.channel = 0;
    interleaveParams.numChannels = 2;

    // Initialize FFT parameters
    fftParams.length = VECTOR_SIZE;
    fftParams.bitReverse = 1;
    fftParams.twiddleTable = msp_cmplx_twiddle_table_256_q15;

    while(keepAppRunning)
    {
        __bis_SR_register(LPM3_bits + GIE);

        // Set the input and buffer pointers
        inputPtr = leaMemoryStartAdd.fftDataParam.input;
        bufferPtr = Audio_getActiveBuffer(&gAudioConfig);

        // Apply Hamming window to buffer samples
        status = msp_mpy_q15(&mpyParams, bufferPtr, leaMemoryStartAdd.fftDataParam.hamming, bufferPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;

        // Zero fill FFT input buffer
        status = msp_cmplx_fill_q15(&fillParams, inputPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;

        // Interleave input samples to real component of FFT input
        status = msp_interleave_q15(&interleaveParams, bufferPtr, inputPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;

        // Reset and start benchmark
        resetBenchmark();
        startBenchmark();

        // Run FFT on complex input
        if (useLea) {
            status = msp_cmplx_fft_fixed_q15(&fftParams, inputPtr);
            if(status != MSP_SUCCESS) P1OUT |= BIT0;
        }
        else {
            FFT_430(inputPtr);
        }

        // Stop benchmark
        stopBenchmark();

#ifndef PWR_BENCHMARK
        uint16_t min;
        uint16_t max;
        uint16_t mag;
        uint16_t scale;
        uint16_t real_abs;
        uint16_t imag_abs;

        // Print cycle counts to string
        if(refreshDisp) {
            sprintf((char *) str, "Cyc:%lu", cycleCounts);
        }

        // Clear the lines by drawing a box
        Graphics_setForegroundColor(&g_sContext, ClrWhite);
        Graphics_fillRectangle(&g_sContext, &clearSpace);
        if(refreshDisp) {
            Graphics_fillRectangle(&g_sContext, &clearSpaceCycleText);
        }
        Graphics_setForegroundColor(&g_sContext, ClrBlack);
        if(refreshDisp) {
            Graphics_drawString(&g_sContext, str,
                                AUTO_STRING_LENGTH, 10, LCD_VERTICAL_MAX-10,
                                GRAPHICS_OPAQUE_TEXT);
            refreshDisp--;
        }

        // Draw the FFT magnitudes but restricting it to only 96 pixel screen
        for(i = 0; i < LCD_HORIZONTAL_MAX; i++) {
            // Approximate magnitude for the positive frequency domain
            real_abs = abs(inputPtr[2*i+0]);
            imag_abs = abs(inputPtr[2*i+1]);
            if(real_abs >= imag_abs) {
                max = real_abs;
                min = imag_abs;
            }
            else {
                max = imag_abs;
                min = real_abs;
            }
            mag = max + 3 * (min >> 3);

            // scale the magnitude down and restrict it to only 83 pixels
            scale = mag >> 6;

            // Saturate scale and draw line
            scale = scale >= (LCD_VERTICAL_MAX-24) ? (LCD_VERTICAL_MAX-25) : scale;
            if(scale != 0)
            {
                // Draw each line only if scale the mag is 0 pixels
                Graphics_drawLineV(&g_sContext, i, LCD_VERTICAL_MAX-13, LCD_VERTICAL_MAX- 13 - scale);
            }

        }

        // Flush this content to display
        Graphics_flushBuffer(&g_sContext);
#endif
    }

    // Deinitialize FFT configuration
    deinitAudio();
}

const _q15 hammingWindow[VECTOR_SIZE] = {
    _Q15(0.0800), _Q15(0.0801), _Q15(0.0806), _Q15(0.0813),
    _Q15(0.0822), _Q15(0.0835), _Q15(0.0850), _Q15(0.0868),
    _Q15(0.0889), _Q15(0.0913), _Q15(0.0939), _Q15(0.0968),
    _Q15(0.1000), _Q15(0.1034), _Q15(0.1071), _Q15(0.1111),
    _Q15(0.1153), _Q15(0.1198), _Q15(0.1245), _Q15(0.1295),
    _Q15(0.1347), _Q15(0.1402), _Q15(0.1459), _Q15(0.1519),
    _Q15(0.1581), _Q15(0.1645), _Q15(0.1712), _Q15(0.1781),
    _Q15(0.1852), _Q15(0.1925), _Q15(0.2001), _Q15(0.2078),
    _Q15(0.2157), _Q15(0.2239), _Q15(0.2322), _Q15(0.2407),
    _Q15(0.2494), _Q15(0.2583), _Q15(0.2673), _Q15(0.2765),
    _Q15(0.2859), _Q15(0.2954), _Q15(0.3051), _Q15(0.3149),
    _Q15(0.3249), _Q15(0.3350), _Q15(0.3452), _Q15(0.3555),
    _Q15(0.3659), _Q15(0.3765), _Q15(0.3871), _Q15(0.3979),
    _Q15(0.4087), _Q15(0.4196), _Q15(0.4305), _Q15(0.4416),
    _Q15(0.4527), _Q15(0.4638), _Q15(0.4750), _Q15(0.4863),
    _Q15(0.4976), _Q15(0.5089), _Q15(0.5202), _Q15(0.5315),
    _Q15(0.5428), _Q15(0.5542), _Q15(0.5655), _Q15(0.5768),
    _Q15(0.5881), _Q15(0.5993), _Q15(0.6106), _Q15(0.6217),
    _Q15(0.6329), _Q15(0.6439), _Q15(0.6549), _Q15(0.6659),
    _Q15(0.6767), _Q15(0.6875), _Q15(0.6982), _Q15(0.7088),
    _Q15(0.7193), _Q15(0.7297), _Q15(0.7400), _Q15(0.7501),
    _Q15(0.7601), _Q15(0.7700), _Q15(0.7797), _Q15(0.7893),
    _Q15(0.7988), _Q15(0.8081), _Q15(0.8172), _Q15(0.8262),
    _Q15(0.8350), _Q15(0.8436), _Q15(0.8520), _Q15(0.8602),
    _Q15(0.8683), _Q15(0.8761), _Q15(0.8837), _Q15(0.8912),
    _Q15(0.8984), _Q15(0.9054), _Q15(0.9121), _Q15(0.9187),
    _Q15(0.9250), _Q15(0.9311), _Q15(0.9369), _Q15(0.9426),
    _Q15(0.9479), _Q15(0.9530), _Q15(0.9579), _Q15(0.9625),
    _Q15(0.9669), _Q15(0.9710), _Q15(0.9748), _Q15(0.9784),
    _Q15(0.9817), _Q15(0.9847), _Q15(0.9875), _Q15(0.9899),
    _Q15(0.9922), _Q15(0.9941), _Q15(0.9958), _Q15(0.9972),
    _Q15(0.9983), _Q15(0.9991), _Q15(0.9997), _Q15(0.9999),
    _Q15(0.9999), _Q15(0.9997), _Q15(0.9991), _Q15(0.9983),
    _Q15(0.9972), _Q15(0.9958), _Q15(0.9941), _Q15(0.9922),
    _Q15(0.9899), _Q15(0.9875), _Q15(0.9847), _Q15(0.9817),
    _Q15(0.9784), _Q15(0.9748), _Q15(0.9710), _Q15(0.9669),
    _Q15(0.9625), _Q15(0.9579), _Q15(0.9530), _Q15(0.9479),
    _Q15(0.9426), _Q15(0.9369), _Q15(0.9311), _Q15(0.9250),
    _Q15(0.9187), _Q15(0.9121), _Q15(0.9054), _Q15(0.8984),
    _Q15(0.8912), _Q15(0.8837), _Q15(0.8761), _Q15(0.8683),
    _Q15(0.8602), _Q15(0.8520), _Q15(0.8436), _Q15(0.8350),
    _Q15(0.8262), _Q15(0.8172), _Q15(0.8081), _Q15(0.7988),
    _Q15(0.7893), _Q15(0.7797), _Q15(0.7700), _Q15(0.7601),
    _Q15(0.7501), _Q15(0.7400), _Q15(0.7297), _Q15(0.7193),
    _Q15(0.7088), _Q15(0.6982), _Q15(0.6875), _Q15(0.6767),
    _Q15(0.6659), _Q15(0.6549), _Q15(0.6439), _Q15(0.6329),
    _Q15(0.6217), _Q15(0.6106), _Q15(0.5993), _Q15(0.5881),
    _Q15(0.5768), _Q15(0.5655), _Q15(0.5542), _Q15(0.5428),
    _Q15(0.5315), _Q15(0.5202), _Q15(0.5089), _Q15(0.4976),
    _Q15(0.4863), _Q15(0.4750), _Q15(0.4638), _Q15(0.4527),
    _Q15(0.4416), _Q15(0.4305), _Q15(0.4196), _Q15(0.4087),
    _Q15(0.3979), _Q15(0.3871), _Q15(0.3765), _Q15(0.3659),
    _Q15(0.3555), _Q15(0.3452), _Q15(0.3350), _Q15(0.3249),
    _Q15(0.3149), _Q15(0.3051), _Q15(0.2954), _Q15(0.2859),
    _Q15(0.2765), _Q15(0.2673), _Q15(0.2583), _Q15(0.2494),
    _Q15(0.2407), _Q15(0.2322), _Q15(0.2239), _Q15(0.2157),
    _Q15(0.2078), _Q15(0.2001), _Q15(0.1925), _Q15(0.1852),
    _Q15(0.1781), _Q15(0.1712), _Q15(0.1645), _Q15(0.1581),
    _Q15(0.1519), _Q15(0.1459), _Q15(0.1402), _Q15(0.1347),
    _Q15(0.1295), _Q15(0.1245), _Q15(0.1198), _Q15(0.1153),
    _Q15(0.1111), _Q15(0.1071), _Q15(0.1034), _Q15(0.1000),
    _Q15(0.0968), _Q15(0.0939), _Q15(0.0913), _Q15(0.0889),
    _Q15(0.0868), _Q15(0.0850), _Q15(0.0835), _Q15(0.0822),
    _Q15(0.0813), _Q15(0.0806), _Q15(0.0801), _Q15(0.0800)
};

#if defined(__IAR_SYSTEMS_ICC__)

// FFT buffer that contains real and imaginary values after being calculated
__persistent int16_t FBuff[VECTOR_SIZE * 2] = {0};

// raw data buffer for processing by the FFT algorithm
__persistent int16_t voice_data[VECTOR_SIZE * 2] = {0};

#elif defined(__TI_COMPILER_VERSION__)

// FFT buffer that contains real and imaginary values after being calculated
#pragma PERSISTENT(FBuff)
int16_t FBuff[VECTOR_SIZE * 2] = {0};

// raw data buffer for processing by the FFT algorithm
#pragma PERSISTENT(voice_data)
int16_t voice_data[VECTOR_SIZE * 2] = {0};

#endif

static const uint8_t bit_rev_index[256] = {
    0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
    0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
    0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
    0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
    0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
    0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
    0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
    0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
    0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
    0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
    0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
    0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
    0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
    0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
    0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
    0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
    0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
    0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
    0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
    0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
    0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
    0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
    0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
    0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
    0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
    0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
    0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
    0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
    0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
    0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
    0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
    0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

void FFT_430(int16_t *input)
{
    uint16_t i;
    uint32_t *srcPtr;
    uint32_t *dstPtr;

    // Perform FFT and store
    FFT_prog(input);

    // Copy FFT result with bit reversal
    srcPtr = (uint32_t *)FBuff;
    dstPtr = (uint32_t *)input;
    for(i = 0; i < VECTOR_SIZE; i++) {
        dstPtr[i] = srcPtr[bit_rev_index[i]];
    }
}
