/* --COPYRIGHT--,BSD
 * Copyright (c) 2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
//******************************************************************************
//!  TI-Design Signal Processing with LEA on MSP430FR5994
//!
//!  William Goh, Brent Peterson, L Reynoso
//!  Texas Instruments Inc.
//!  Sept 2019
//******************************************************************************

#include "application.h"

//******************************************************************************
// Globals
//******************************************************************************
// Allocate DSPLib structures
static msp_status status;
static msp_fill_q15_params fillParams;
static msp_shift_q15_params shiftParams;
static msp_offset_q15_params offsetParams;
static msp_biquad_cascade_df2_q15_params iirParams;

 //Allocating DMA1 to copy data from ADC buffer to LEA
 static DMA_initParam dma1Config;

static void copyCoeffs(void *coeffs, uint16_t length)
{
    // Using DMA1 to copy the coefficients to LEA memory first
    dma1Config.channelSelect = DMA_CHANNEL_1;
    dma1Config.transferModeSelect = DMA_TRANSFER_REPEATED_BURSTBLOCK;
    dma1Config.transferSize = length/2;
    dma1Config.triggerSourceSelect = (DMA1TSEL__DMAREQ >> 8);
    dma1Config.transferUnitSelect = DMA_SIZE_SRCWORD_DSTWORD;
    dma1Config.triggerTypeSelect = DMA_TRIGGER_RISINGEDGE;

    // Initialize DMA1 with the config
    DMA_init(&dma1Config);

    // Set source address to the coefficient constants in memory
    DMA_setSrcAddress(DMA_CHANNEL_1,
                      (uint32_t)coeffs,
                      DMA_DIRECTION_INCREMENT);

    // Set the destination to LEA memory
    DMA_setDstAddress(DMA_CHANNEL_1,
                      (uint32_t)leaMemoryStartAdd.iirDataParam.coeffs,
                      DMA_DIRECTION_INCREMENT);

    // Enable DMA channel 1 interrupt
    DMA_enableInterrupt(DMA_CHANNEL_1);

    // Re-enable the DMA transfer
    DMA_enableTransfers(DMA_CHANNEL_1);

    // Start the data transfer
    DMA_startTransfer(DMA_CHANNEL_1);

    // Enter LPM and wait for transfer to complete
    __bis_SR_register(LPM0_bits + GIE);
}

static void initAudio(void)
{
    // Initialize the microphone for recording
    gAudioConfig.audioBuffer1 = (int16_t *)&leaMemoryStartAdd.iirDataParam.input[0];
    gAudioConfig.audioBuffer2 = (int16_t *)&leaMemoryStartAdd.iirDataParam.input[IIR_SAMPLES];
    gAudioConfig.bufferSize = IIR_SAMPLES;
    gAudioConfig.sampleRate = IIR_SAMPLING_FREQUENCY;
    Audio_setupCollect(&gAudioConfig);

    // Reset the playback counter
    dacBufferPlaybackPointer = 0;

    // Initialize the audio playback config. This also reconfigures the
    // SPI configuration
    gPlaybackConfig.playbackBuffer1 = (int16_t *)&leaMemoryStartAdd.iirDataParam.result[0];
    gPlaybackConfig.playbackBuffer2 = (int16_t *)&leaMemoryStartAdd.iirDataParam.result[IIR_SAMPLES];
    gPlaybackConfig.bufferSize = IIR_SAMPLES;
    gPlaybackConfig.sampleRate = IIR_SAMPLING_FREQUENCY;
    Audio_setupPlayback(&gPlaybackConfig);

    // Start the recording by enabling the timer
    Audio_startCollect(&gAudioConfig);
}

static void deinitAudio(void)
{
    // Stop the audio playback
    Audio_stopPlayback(&gPlaybackConfig);

    // Stop the audio collection
    Audio_stopCollect(&gAudioConfig);
    Audio_shutdownCollect(&gAudioConfig);
}

void runIirFilter(void)
{
    int16_t *inputPtr;
    int16_t *resultPtr;
    uint16_t iirStages;

    keepAppRunning = true;

    // Clear the display first
    Graphics_clearDisplay(&g_sContext);

    // Create text for display
    Graphics_drawStringCentered(&g_sContext, "Real-Time IIR",
                                AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-66,
                                GRAPHICS_OPAQUE_TEXT);

    Graphics_drawStringCentered(&g_sContext, "Filtering",
                                AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-56,
                                GRAPHICS_OPAQUE_TEXT);

    switch(iirAppMode)
    {
    case IIR_LOW_PASS_1000:
        Graphics_drawStringCentered(&g_sContext, "Low-Pass @ 1kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirLowPass1kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirLowPass1kHz, sizeof(iirLowPass1kHz));
        break;
    case IIR_LOW_PASS_2000:
        Graphics_drawStringCentered(&g_sContext, "Low-Pass @ 2kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirLowPass2kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirLowPass2kHz, sizeof(iirLowPass2kHz));
        break;
    case IIR_LOW_PASS_3000:
        Graphics_drawStringCentered(&g_sContext, "Low-Pass @ 3kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirLowPass3kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirLowPass3kHz, sizeof(iirLowPass3kHz));
        break;
    case IIR_LOW_PASS_4000:
        Graphics_drawStringCentered(&g_sContext, "Low-Pass @ 4kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirLowPass4kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirLowPass4kHz, sizeof(iirLowPass4kHz));
        break;
    case IIR_HIGH_PASS_1000:
        Graphics_drawStringCentered(&g_sContext, "High-Pass @ 1kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirHighPass1kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirHighPass1kHz, sizeof(iirHighPass1kHz));
        break;
    case IIR_HIGH_PASS_2000:
        Graphics_drawStringCentered(&g_sContext, "High-Pass @ 2kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirHighPass2kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirHighPass2kHz, sizeof(iirHighPass2kHz));
        break;
    case IIR_HIGH_PASS_3000:
        Graphics_drawStringCentered(&g_sContext, "High-Pass @ 3kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirHighPass3kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirHighPass3kHz, sizeof(iirHighPass3kHz));
        break;
    case IIR_HIGH_PASS_4000:
        Graphics_drawStringCentered(&g_sContext, "High-Pass @ 4kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirHighPass4kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirHighPass4kHz, sizeof(iirHighPass4kHz));
        break;
    case IIR_BAND_PASS_1000:
        Graphics_drawStringCentered(&g_sContext, "Band-Pass @ 1kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandPass1kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandPass1kHz, sizeof(iirBandPass1kHz));
        break;
    case IIR_BAND_PASS_2000:
        Graphics_drawStringCentered(&g_sContext, "Band-Pass @ 2kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandPass2kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandPass2kHz, sizeof(iirBandPass2kHz));
        break;
    case IIR_BAND_PASS_3000:
        Graphics_drawStringCentered(&g_sContext, "Band-Pass @ 3kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandPass3kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandPass3kHz, sizeof(iirBandPass3kHz));
        break;
    case IIR_BAND_PASS_4000:
        Graphics_drawStringCentered(&g_sContext, "Band-Pass @ 4kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandPass4kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandPass4kHz, sizeof(iirBandPass4kHz));
        break;
    case IIR_BAND_STOP_1000:
        Graphics_drawStringCentered(&g_sContext, "Band-Stop @ 1kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandStop1kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandStop1kHz, sizeof(iirBandStop1kHz));
        break;
    case IIR_BAND_STOP_2000:
        Graphics_drawStringCentered(&g_sContext, "Band-Stop @ 2kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandStop2kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandStop2kHz, sizeof(iirBandStop2kHz));
        break;
    case IIR_BAND_STOP_3000:
        Graphics_drawStringCentered(&g_sContext, "Band-Stop @ 3kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandStop3kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandStop3kHz, sizeof(iirBandStop3kHz));
        break;
    case IIR_BAND_STOP_4000:
        Graphics_drawStringCentered(&g_sContext, "Band-Stop @ 4kHz",
                                    AUTO_STRING_LENGTH, LCD_HORIZONTAL_MAX/2, LCD_HORIZONTAL_MAX-46,
                                    GRAPHICS_OPAQUE_TEXT);
        iirStages = sizeof(iirBandStop4kHz)/sizeof(msp_biquad_df2_q15_coeffs);
        copyCoeffs((void *)iirBandStop4kHz, sizeof(iirBandStop4kHz));
        break;
    default: break;
    }

    // Flush this content to display
    Graphics_flushBuffer(&g_sContext);

    // Initialize audio configuration
    initAudio();

    // Initialize shift parameters to scale input to Q13
    shiftParams.length = IIR_SAMPLES;
    shiftParams.shift = -2;

    // Initialize offset parameters to add DC bias before playback
    offsetParams.length = IIR_SAMPLES;
    offsetParams.offset = 0x2000;

    // Zero initialize IIR states
    fillParams.length = sizeof(leaMemoryStartAdd.iirDataParam.states)/sizeof(_q15);
    fillParams.value = 0x0000;
    status = msp_fill_q15(&fillParams, (_q15 *)leaMemoryStartAdd.iirDataParam.states);

    // Initialize IIR parameters
    iirParams.length = IIR_SAMPLES;
    iirParams.stages = iirStages;
    iirParams.coeffs = leaMemoryStartAdd.iirDataParam.coeffs;
    iirParams.states = leaMemoryStartAdd.iirDataParam.states;

    while(keepAppRunning)
    {
        __bis_SR_register(LPM3_bits + GIE);

        // Switch the audio playback buffer here
        Audio_switchPlaybackBuffer(&gPlaybackConfig);

        // Start the audio playback here for critical timing reasons
        // as the audio needs to always start from the beginning
        Audio_startPlayback(&gPlaybackConfig);

        // Get the input and playback buffers
        inputPtr = Audio_getActiveBuffer(&gAudioConfig);
        resultPtr = Audio_getEmptyPlaybackBuffer(&gPlaybackConfig);

        // Scale input samples to Q13
        status = msp_shift_q15(&shiftParams, inputPtr, inputPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;

        // Run IIR filter
        status = msp_biquad_cascade_df2_q15(&iirParams, inputPtr, resultPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;

        // Add DC offset before playback
        status = msp_offset_q15(&offsetParams, resultPtr, resultPtr);
        if(status != MSP_SUCCESS) P1OUT |= BIT0;
    }

    // Deinitialize IIR configuration
    deinitAudio();
}
